
/*
 * COMMANDS.C
 *
 *  Global Routines:    DO_QUIT()
 *                      DO_EXIT() 
 *                      DO_CD()
 *                      DO_ECHO()
 *                      DO_GO()
 *                      DO_SOURCE()
 *                      DO_SHELL()
 *                      DO_WRITE()
 *                      DO_WHEAD()
 *                      DO_PIPE()
 *                      DO_DELNEXT()
 *                      DO_NUMBER()
 *                      DO_NEXT()
 *                      DO_HEADER()
 *                      DO_TYPE()
 *                      DO_DELETE()
 *                      DO_UNDELETE()
 *                      DO_MARK()
 *		        DO_BREAK()
 *			DO_ORDER()
 *
 *  Static Routines:    None.
 *
 */

#include <stdio.h>
#include <sys/file.h>
#include "dmail.h"

#define LAST_TYPE   0
#define LAST_HEADER 1

int Last_operation;
int Last_deleted = -1;

Prototype int do_quit(char *garbage, int com);
Prototype int do_exit(char *garbage, int com);
Prototype int do_cd(char *garbage, int com);
Prototype int do_echo(char *str, int com);
Prototype int do_go(char *garbage, int com);
Prototype int do_source(char *scratch, int overide);
Prototype int do_shell(char *str, int com);
Prototype int do_write(char *garbage, int com);
Prototype int do_whead(char *garbage, int com);
Prototype int do_pipe(char *garbage, int com);
Prototype int do_delprev(char *garbage, int com);
Prototype int do_delnext(char *garbage, int com);
Prototype int do_order(char *str, int com);
Prototype int do_number(char *str, int com);
Prototype int do_next(char *str, int com);
Prototype int do_header(char *garbage, int com);
Prototype int do_type(char *garbage, int com);
Prototype int do_mark(char *garbage, int mask);
Prototype int do_unmark(char *garbage, int mask);
Prototype int do_break(char *garbage, int mask);
Prototype int do_ver(char *garbage, int mask);


int
do_quit(char *garbage, int com)
{
    int fd, r, back;
    char *str;

    push_break();
    if (get_inode (mail_file) == get_inode (output_file)) {
        back = save_file (0, 0, ST_DELETED | ST_STORED);
    } else {
        r = write_file (output_file, O_CREAT, ST_READ, ST_DELETED | ST_STORED);
        if (r < 0) {
            printf ("Unable to write to %s\n", output_file);
            back = save_file (0, 0, ST_DELETED | ST_STORED);
        } else {
            back = save_file (0, 0, ST_READ | ST_DELETED | ST_STORED);
        }
    }
    if (back < 0)
        printf ("Unable to update %s\n", mail_file);
    if (back > 0) 
        printf ("%d  kept in %s\n", back, mail_file);
    sleep (1);
    if ((fd = open (mail_file, O_RDONLY, 0)) >= 0) {
        read (fd, Buf, 1);
        close (fd);
    }
    if (!com) 
        done (0);
    free_entry();
    if (av[1] == 0) {
        if (!Silence && !S_quiet)
            puts ("NO FROM FILE SPECIFIED");
        av[1] = mail_file;
        av[2] = NULL;
    }
    Did_cd = 0;
    mail_file = realloc (mail_file, strlen(av[1]) + 1);
    strcpy (mail_file, av[1]);
    str = (av[2]) ? av[2] : mail_file;
    output_file = realloc (output_file, strlen(str) + 1);
    strcpy (output_file, str);
    initial_load_mail();
    m_select (Nulav, M_RESET, 0);
    pop_break();
    if (!Silence && !S_quiet)
        printf ("\nRF %-20s   WF %-20s\n", mail_file, output_file);
    return(1);
}


int
do_exit(char *garbage, int com)
{
    char *str;

    if (!com)
        done (0);
    push_break();
    free_entry();
    if (av[1] == 0) {
        if (!Silence && !S_quiet)
            puts ("NO FROM FILE SPECIFIED");
        av[1] = mail_file;
        av[2] = NULL;
    }
    mail_file = realloc (mail_file, strlen(av[1]) + 1);
    strcpy (mail_file, av[1]);
    str = (av[2]) ? av[2] : mail_file;
    output_file = realloc (output_file, strlen(str) + 1);
    strcpy (output_file, str);
    initial_load_mail();
    m_select (Nulav, M_RESET, 0);
    pop_break();
    if (!Silence && !S_quiet)
        printf ("\nRF %-20s   WF %-20s\n", mail_file, output_file);
    return(1);
}

int
do_cd(char *garbage, int com)
{
    char *dir = (ac < 2) ? home_dir : av[1];

    if (chdir(dir) < 0) {
        printf ("Cannot CD to %s\n", dir);
        return (-1);
    }
    ++Did_cd;
    return (1);
}

int
do_echo(char *str, int com)
{
    puts (next_word(str));
    fflush(stdout);
    return (1);
}

int
do_go(char *garbage, int com)
{
    int i;

    if (ac < 2) {
        puts ("go to which article?");
    }
    rewind_range (1);
    i = get_range();
    if (i < 0) {
        if (!Silence && !S_quiet)
            printf ("Message #%d does not exist\n", i);
        return (-1);
    }
    if (i == 0) {
        if (!Silence && !S_quiet)
            puts ("No Message");
        return (-1);
    }
    Current = indexof(i);
    return (1);
}

int
do_source(char *scratch, int overide)
{
    char comline[1024];
    FILE *volatile fi = NULL;

    if (ac < 2) {
        puts ("No file argument to source");
        return (-1);
    }
    if (push_base()) {
        push_break();
        pop_base();
        if (fi != NULL)
            fclose ((FILE *)fi);
        pop_break();
        return (-1);
    }
    push_break();
    fi = fopen (av[1], "r");
    pop_break();
    if (fi == NULL) {
        if (!overide)
            printf ("Cannot open %s\n", av[1]);
        return (-1);
    }
    while (fgets (comline, 1024, (FILE *)fi) != NULL) {
        comline[strlen(comline) - 1] = '\0';
    if (comline[0] != '#')
            exec_command (comline);
    }
    push_break();
    fclose ((FILE *)fi);
    fi = NULL;
    pop_break();
    pop_base();
    return (1);
}

int
do_shell(char * volatile str, int com)
{
    volatile int pid, ret;
    char * volatile shell;
    char * volatile ptr;
    char * volatile args;

    shell = getenv("SHELL");
    if (shell == NULL)
        shell = "/bin/sh";
    ptr = shell + (strlen(shell)-1);
    while (ptr > shell && *ptr && *ptr != '/')
        --ptr;
    args = (strcmp(ptr, "/sh"))? "-fc" : "-c";
    push_break();
    str = next_word (str);
    if (strlen (str)) {
        if ((pid = vfork()) == 0) {
            execl (shell, shell, args, str, NULL);
            _exit (1);
        }
    } else {
        if ((pid = vfork()) == 0) {
            execl (shell, shell, NULL);
            _exit (1);
        }
    }
    while ((ret = wait(0)) > 0) {
        if (ret == pid)
            break;
    }
    pop_break();
    return (1);
}


int
do_write(char *garbage, int com)
{
    char *file;
    int r, count = 0;
    register int i, j;

    if (ac < 2) {
        puts ("You must specify at least a file-name");
        return (-1);
    }
    file = av[1];
    rewind_range (2);
    push_break();
    while ((i = get_range()) != 0) {
        j = indexof (i);
        if (j >= 0  &&  !(Entry[j].status & ST_DELETED)) {
            Entry[j].status |= ST_STORED | ST_SCR;
            ++count;
        }
    }
    r = write_file (file, O_CREAT, ST_SCR, 0);
    rewind_range (2);
    if (r > 0) {
        while ((i = get_range()) != 0) {
            j = indexof (i);
            if (j >= 0)
                Entry[j].status &= ~ST_SCR;
        }
        if (!Silence && !S_quiet)
            printf ("%d Items written\n", count);
    } else {
        while ((i = get_range()) != 0) {
            j = indexof (i);
            if (j >= 0)
                Entry[j].status &= ~(ST_SCR | ST_STORED);
        }
        printf ("Could not write to file %s\n", file);
    }
    pop_break();
    return (1);
}

int
do_whead(char *garbage, int com)
{
    FILE *fo;
    int i;

    if (ac < 3) {
        puts("You must specify a file-name and at least one header");
        return (-1);
    }
    if (Current < 0 || Current >= Entries) {
        puts("whead: no message");
        return (-1);
    }
    push_break();
    if (strcmp(av[1], "-") == 0)
	fo = stdout;
    else
	fo = fopen(av[1], "a");
    if (fo == NULL) {
        printf ("Could not create file %s\n", av[1]);
	return(-1);
    }
    for (i = 2; av[i]; ++i) {
	int len;
	const char *str = get_field(av[i], &len);
	if (str) {
	    fwrite(str, len, 1, fo);
	    fputc('\n', fo);
	}
    }
    if (fo != stdout)
	fclose(fo);
    pop_break();
    return (1);
}

int
do_pipe(char *garbage, int com)
{
    char *prog;
    int r, count = 0;
    register int i, j;

    if (ac < 2) {
        puts ("You must specify a command to pipe the mail through");
        return (-1);
    }
    prog = av[1];
    rewind_range (2);
    push_break();
    while ((i = get_range()) != 0) {
        j = indexof (i);
        if (j >= 0  &&  !(Entry[j].status & ST_DELETED)) {
            Entry[j].status |= ST_STORED | ST_SCR;
            ++count;
        }
    }
    r = write_prog (prog, O_CREAT, ST_SCR, 0);
    rewind_range (2);
    if (r > 0) {
        while ((i = get_range()) != 0) {
            j = indexof (i);
            if (j >= 0)
                Entry[j].status &= ~ST_SCR;
        }
#if 0
        if (!Silence && !S_quiet)
            printf ("%d Items written\n", count);
#endif
    } else {
        while ((i = get_range()) != 0) {
            j = indexof (i);
            if (j >= 0)
                Entry[j].status &= ~(ST_SCR | ST_STORED);
        }
        printf ("Could not write to program %s\n", prog);
    }
    pop_break();
    return (1);
}


/*
 * DB, added 3 Oct 1988
 */

int
do_delprev(char *garbage, int com)
{
    do_mark("", ST_DELETED);
    if (Current)
	return(do_next("", -1));
    return(-1);
}

int
do_delnext(char *garbage, int com)
{
    static int warning;

    if (!warning  &&  Last_operation == LAST_HEADER) {
        ++warning;
        puts ("Note that the next command is displaying headers only at");
        puts ("this point.  (one-time warning, NOTHING deleted");
        return (-1);
    }
    if (do_mark("", ST_DELETED) > 0)
        return (do_next("", 1));
    return (-1);
}

/*
 * do_order()
 *
 */

int 
do_order(char *str, int com)
{
    int i;

    rewind_range (1);
    while ((i = get_range()) != 0) {
	struct ENTRY t;

        i = indexof(i);
	t = Entry[i];
	t.no = 0;	/* deselect */

	if (i == Current)
	    ++Current;

	bcopy(&Entry[0], &Entry[1], i * sizeof(struct ENTRY));
	Entry[0] = t;
	ChangesMade = 1;
    }
    if (Current >= Entries)
	Current = Entries - 1;
    return(1);
}

int
do_number(char *str, int com)
{
    int x;

    x = indexof (atoi(str));
    if (x < 0) {
        puts ("Non existant message");
        return (-1);
    }
    Current = x;
    switch (Last_operation) {
    case LAST_TYPE:
        return (do_type(NULL, 0));
    case LAST_HEADER:
        return (do_header(NULL, 0));
    default:
        puts ("Internal Error NEXT");
        return (-1);
    }
}

int
do_next(char *str, int com)
{
    int ok;

    push_break();
    if (com > 0) {
        if (++Current > Entries)
            Current = Entries;
        if (fix() < 0) {
            puts ("End of file");
            pop_break();
            return (-1);
        }
        --com;
    }
    if (com < 0) {
        ++com;
        ok = 0;
        while (--Current >= 0) {
            if (Entry[Current].no  &&  !(Entry[Current].status & ST_DELETED)) {
                ok = 1;
                break;
            }
        }
        if (!ok) {
            puts ("Start of file");
        Current = 0;
            fix();
            pop_break();
            return (-1);
        }
    }
    pop_break();
    if (!com) {
        switch (Last_operation) {
        case LAST_TYPE:
            return (do_type(NULL, 0));
        case LAST_HEADER:
            return (do_header(NULL, 0));
        }
    }
    return (1);
}

int
do_header(char *garbage, int com)
{
    struct ENTRY *en;

    Last_operation = LAST_HEADER;
    if (push_base()) {
        push_break();
        pop_base();
        PAGER ((void *)-1);
        fflush (stdout);
        pop_break();
        return (-1);
    }
    if (single_position() < 0)
        return (-1);
    en = &Entry[Current];
    if (Current < 0) {
	puts("Software error #commands.0");
	exit(1);
    }
    PAGER (NULL);
    sprintf (Puf, "MESSAGE HEADER #%d (%d) %s\n",
	en->no,
	Current + 1,
	((en->status & ST_DELETED) ? "  DELETED" : "")
    );
    PAGER (Puf);

    WPAGER(FileData + en->fpos, en->hlen);
    PAGER("");

    pop_base();
    PAGER ((void *)-1);
    return(1);
}

int
do_type(char *garbage, int com)
{
    struct ENTRY *en;
    int i;

    Last_operation = LAST_TYPE;
    if (push_base()) {
        push_break();
        pop_base();
        PAGER ((void *)-1);
        fflush (stdout);
        pop_break();
        return (-1);
    }
    if (single_position() < 0)
        return (-1);
    en = &Entry[Current];
    if (Current < 0) {
	puts("Software Error #commands.1");
	exit(1);
    }

    PAGER (NULL);
    sprintf (Puf, "MESSAGE TEXT #%d (%d) %s\n", 
	en->no,
	Current + 1,
	((en->status & ST_DELETED) ? "  DELETED" : "")
    );
    PAGER (Puf);
    for (i = 0; i < Listsize; ++i) {
        if (*Entry[Current].fields[header[i]] && *Entry[Current].fields[header[i]] != '\n') {
            sprintf (Puf, "%-10s", Find[header[i]].search);
            FPAGER (Puf);
	    WPAGER(en->fields[header[i]], data_line_length(en->fields[header[i]], 1));
        }
    }
    PAGER ("");

    WPAGER(FileData + en->fpos + en->hlen, en->tlen - en->hlen);

    en->status |= ST_READ;
    PAGER ((void *)-1);
    pop_base();
    return (1);
}

int
do_mark(char *garbage, int mask)
{
    int count = 0;
    register int i, j;

    rewind_range (1);
    push_break();
    while ((i = get_range()) != 0) {
        j = indexof (i);
        if (j >= 0) {
            if (mask & ST_DELETED)
                Last_deleted = j;
            if ((Entry[j].status & mask) != mask) {
                Entry[j].status |= mask;
                if (Entry[j].status & ST_DELETED)
                    Entry[j].status &= ~(ST_STORED | ST_READ | ST_TAG);
                ++count;
            }
        }
    }
    if (!Silence && !S_quiet)
        printf ("%d  Items\n", count);
    pop_break();
    return (1);
}

int
do_unmark(char *garbage, int mask)
{
    int count = 0;
    register int i, j;
    register struct ENTRY *en;

    push_break();
    if (ac == 1 && (mask & ST_DELETED) && Last_deleted != -1)  {
        en = &Entry[Last_deleted];
        if (en->no) {
            en->status &= ~mask;
            printf ("Undeleted last deleted message (# %d)\n", en->no);
            Current = Last_deleted;
            Last_deleted = -1;
        } else {
            puts ("Last deleted message not within current select bounds");
            pop_break();
            return (-1);
        }
        pop_break();
        return (1);
    }
    rewind_range (1);
    while ((i = get_range()) != 0) {
        j = indexof (i);
        if (j >= 0) {
            if (Entry[j].status & mask) {
                Entry[j].status &= ~mask;
                ++count;
            }
        }
    }
    if (!Silence && !S_quiet)
        printf ("%d  Items\n", count);
    pop_break();
    return(1);	/* was based on count, now is not */
}

int
do_break(char *garbage, int mask)
{
    if (mask)
        pop_break();
    else
        push_break();
    return (1);
}

int
do_ver(char *garbage, int mask)
{
    puts (VERSION);
    return(1);
}

